/*
 *                            COPYRIGHT
 *
 *  cschem - modular/flexible schematics editor - libcschem (core library)
 *  Copyright (C) 2018,2019 Tibor 'Igor2' Palinkas
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.*
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/cschem
 *    lead developer: email to cschem (at) igor2.repo.hu
 *    mailing list: cschem (at) list.repo.hu (send "subscribe")
 */

#include "config.h"

#include <stdio.h>

#include "abstract.h"
#include "actions_csch.h"
#include "concrete.h"
#include "cnc_line.h"
#include "cnc_arc.h"
#include "cnc_poly.h"
#include "cnc_bitmap.h"
#include "cnc_conn.h"
#include "cnc_grp.h"
#include "cnc_pen.h"
#include "engine.h"
#include "message.h"
#include "project.h"

static int compile_attributes(csch_ahdr_t *dst, const csch_attribs_t *src)
{
	htsp_entry_t *e;
	for(e = htsp_first(src); e; e = htsp_next(src, e)) {
		const csch_attrib_t *a = e->value;
		csch_source_arg_t src;
		
		src.type = 'c';
		src.source = (a->source.used > 0) ? a->source.array[0] : NULL;
		src.desc = NULL;
		if (a->val != NULL)
			csch_attrib_set(&dst->attr, a->prio, a->key, a->val, src, NULL);
		else
			csch_attrib_set_arr(&dst->attr, a->prio, a->key, &a->arr, src, NULL);
	}
	return 0;
}

static int compile_wire_net(csch_abstract_t *dst, const csch_sheet_t *sheet, csch_cgrp_t *src)
{
	char tmpname[128];
	const char *name;
	const csch_attrib_t *aname;
	csch_anet_t *net;

	aname = csch_attrib_get(&src->hdr.attr, "name");
	if ((aname == NULL) || (aname->key == NULL) || (*aname->key == '\0')) {
		TODO("needs better naming");
		sprintf(tmpname, "anon_%p", (void *)src);
		name = tmpname;
	}
	else
		name = aname->key;

	net = csch_anet_get(dst, name, 1);
	if (net == NULL)
		return -1;

	vtl0_append(&src->aid, net->hdr.aid);

	return compile_attributes(&net->hdr, &src->hdr.attr);
}

static int compile_symbol(csch_abstract_t *dst, const csch_sheet_t *sheet, csch_cgrp_t *src)
{
	char tmpname[128];
	htip_entry_t *e;
	const char *name, *pname_tmp;
	const csch_attrib_t *aname = NULL;
	csch_acomp_t *comp;
	int res = 0;

	if (src->hdr.type == CSCH_CTYPE_GRP_REF)
		aname = csch_cgrp_ref_get_attr(src, "name");
	else
		aname = csch_attrib_get(&src->hdr.attr, "name");

	if ((aname == NULL) || (aname->key == NULL) || (*aname->key == '\0')) {
		TODO("needs better naming");
		sprintf(tmpname, "anon_%p", (void *)src);
		name = tmpname;
	}
	else
		name = aname->val;

	comp = csch_acomp_get(dst, name, 1);
	if (comp == NULL)
		return -1;

	vtl0_append(&src->aid, comp->hdr.aid);

	res |= compile_attributes(&comp->hdr, &src->hdr.attr);
	if ((src->hdr.type == CSCH_CTYPE_GRP_REF) && (src->data.ref.grp != NULL))
		res |= compile_attributes(&comp->hdr, &src->data.ref.grp->hdr.attr);

	csch_eng_call(sheet->parent, CSCH_ENGHK_SYMBOL_JOINED_COMPONENT,
		FGW_COBJ, src, FGW_AOBJ, comp, FGW_INVALID);

	/* compile ports */
	for(e = htip_first(&src->id2obj); e != NULL; e = htip_next(&src->id2obj, e)) {
		csch_cgrp_t *t = e->value;
		const csch_attrib_t *trole, *tname;
		const char *pname;
		csch_aport_t *port;

		if ((t->hdr.type != CSCH_CTYPE_GRP) && (t->hdr.type != CSCH_CTYPE_GRP_REF))
			continue;

		if (t->hdr.type == CSCH_CTYPE_GRP_REF)
			trole = csch_cgrp_ref_get_attr(t, "role");
		else
			trole = csch_attrib_get(&t->hdr.attr, "role");
		if ((trole == NULL) || (trole->val == NULL) || (strcmp(trole->val, "terminal") != 0))
			continue;

		if (t->hdr.type == CSCH_CTYPE_GRP_REF)
			tname = csch_cgrp_ref_get_attr(t, "name");
		else
			tname = csch_attrib_get(&t->hdr.attr, "name");

		if ((tname == NULL) || (tname->key == NULL) || (*tname->key == '\0')) {
			TODO("needs better naming");
			sprintf(tmpname, "anon_%p", (void *)src);
			pname = tmpname;
		}
		else
			pname = tname->val;

		csch_eng_call_strmod(sheet->parent, CSCH_ENGHK_TERMINAL_NAME_TO_PORT_NAME,
			&pname_tmp, pname, FGW_AOBJ, comp, FGW_COBJ, t, FGW_INVALID);

		port = csch_aport_get(dst, comp, pname_tmp, 1);
		port->parent = comp;
		vtl0_append(&t->aid, port->hdr.aid);

		csch_eng_free_strmod(&pname_tmp, pname);
	}
	return res;
}

static int compile_conn(csch_abstract_t *dst, const csch_sheet_t *sheet, const csch_conn_t *src)
{
	long n;
	csch_anet_t *net = NULL;

	if (src->conn.used < 2) {
		csch_error("Ignoring conn object with %d connection(s) - needs at least 2\n", src->conn.used);
		return -1;
	}


	/* first pass: find the net */
	for(n = 0; n < src->conn.used; n++) {
		const csch_chdr_t *obj = src->conn.array[n];
		const csch_cgrp_t *grp = obj->parent;
		csch_ahdr_t *a;

		TODO("remove this:");
		if (grp == NULL) continue;

		if ((grp->hdr.type != CSCH_CTYPE_GRP) && (grp->hdr.type != CSCH_CTYPE_GRP_REF))
			continue;

		if (grp->aid.used < 1) {
			csch_error("Invalid object on the conn list (empty)\n");
			return -1;
		}

		a = htip_get(&dst->aid2obj, grp->aid.array[0]);
		if (a == NULL) {
			csch_error("Invalid object on the conn list (NULL)\n");
			return -1;
		}
		switch(a->type) {
			case CSCH_ATYPE_NET:
				if (net != NULL) {
					TODO("figure how to handle network collisions");
					csch_error("Network collision\n");
					return -1;
				}
				net = (csch_anet_t *)a;
				break;
			case CSCH_ATYPE_PORT:
				break;
			default:
				csch_error("Invalid object on the conn list (type)\n");
				return -1;
		}
		printf(" aid=%ld (%ld) %p\n", grp->aid.array[0], (long)grp->aid.used, (void *)a);
	}

	if (net == NULL) {
		TODO("this should work when two terminals are directly connected: allocate a disposable net (because later on a real net may also connect: if the terminal is attribute-connected)");
		csch_error("Invalid object on the conn list (no net)\n");
		return -1;
	}


	/* second pass: make the connections */
	for(n = 0; n < src->conn.used; n++) {
		const csch_chdr_t *obj = src->conn.array[n];
		const csch_cgrp_t *grp = obj->parent;
		csch_ahdr_t *a;

		TODO("remove this:");
		if (grp == NULL) continue;

		if ((grp->hdr.type != CSCH_CTYPE_GRP) && (grp->hdr.type != CSCH_CTYPE_GRP_REF))
			continue;

		if (grp->aid.used < 1)
			continue;

		a = htip_get(&dst->aid2obj, grp->aid.array[0]);
		if (a == &net->hdr)
			continue;

		vtp0_append(&net->conns, a);
	}
	return 0;
}

int csch_compile_sheet(csch_abstract_t *dst, const csch_sheet_t *src)
{
	int res = 0;
	const csch_chdr_t *h;

	/* reset aid */
	for(h = gdl_first(&src->active); h != NULL; h = gdl_next(&src->active, h)) {
		if (csch_obj_is_grp(h)) {
			csch_cgrp_t *grp = (csch_cgrp_t *)h;
			grp->aid.used = 0;
		}
	}

	/* compile nets and components */
	for(h = gdl_first(&src->active); h != NULL; h = gdl_next(&src->active, h)) {
		csch_cgrp_t *grp = (csch_cgrp_t *)h;

		if (h->indirect)
			continue;

		if ((h->type != CSCH_CTYPE_GRP) && (h->type != CSCH_CTYPE_GRP_REF))
			continue;

		/* can use enum role here because role attribute overwrite is handled in csch_cgrp_ref_render() */
		switch(grp->role) {
			case CSCH_ROLE_WIRE_NET:   res |= compile_wire_net(dst, src, grp); break;
			case CSCH_ROLE_SYMBOL:     res |= compile_symbol(dst, src, grp); break;

			case CSCH_ROLE_invalid:
			case CSCH_ROLE_empty:
			case CSCH_ROLE_BUS_NET:
			case CSCH_ROLE_BUS_TERMINAL:
			case CSCH_ROLE_HUB_POINT:
			case CSCH_ROLE_TERMINAL:
			case CSCH_ROLE_JUNCTION:
				break;
		}
	}

	/* compile connections between nets and components (the "netlist" part) */
	for(h = gdl_first(&src->active); h != NULL; h = gdl_next(&src->active, h)) {
		const csch_conn_t *conn = (const csch_conn_t *)h;

		if (h->indirect)
			continue;

		if (h->type != CSCH_CTYPE_CONN)
			continue;
		res |= compile_conn(dst, src, conn);
	}

	return res;
}

int csch_compile_post(csch_project_t *proj, csch_abstract_t *dst)
{
	htsp_entry_t *e, *p;

	for(e = htsp_first(&dst->comps); e != NULL; e = htsp_next(&dst->comps, e)) {
		csch_acomp_t *comp = e->value;
		csch_eng_call(proj, CSCH_ENGHK_COMPILE_COMPONENT1, FGW_AOBJ, comp, FGW_INVALID);
		for(p = htsp_first(&comp->ports); p != NULL; p = htsp_next(&comp->ports, p))
			csch_eng_call(proj, CSCH_ENGHK_COMPILE_PORT, FGW_AOBJ, p->value, FGW_INVALID);
		csch_eng_call(proj, CSCH_ENGHK_COMPILE_COMPONENT2, FGW_AOBJ, comp, FGW_INVALID);
	}

	for(e = htsp_first(&dst->nets); e != NULL; e = htsp_next(&dst->nets, e))
		csch_eng_call(proj, CSCH_ENGHK_COMPILE_NET, FGW_AOBJ, e->value, FGW_INVALID);
	return 0;
}

int csch_compile_project(csch_project_t *prj, csch_abstract_t *dst, int quiet)
{
	int res = 0, r;
	long n;

	for(n = 0; n < vtp0_len(&prj->sheets); n++) {
		csch_sheet_t *sheet = *vtp0_get(&prj->sheets, n, 0);

		r = csch_compile_sheet(dst, sheet);
		if (r != 0)
			csch_message(CSCH_MSG_ERROR, "sheet #%d (%s) compilation failed: %d\n", n, sheet->hidlib.filename, r);
		res |= r;
	}
	csch_compile_post(prj, dst);
	return res;
}
